package submit.aiwolf.wordwolf;

import java.io.IOException;
import java.io.InputStreamReader;
import java.util.ArrayList;
import java.util.Deque;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Objects;
import java.util.Set;

import org.aiwolf.client.lib.*;
import org.aiwolf.common.data.*;
import org.aiwolf.common.net.GameInfo;
import org.aiwolf.common.net.GameSetting;

public abstract class DemoBasePlayer implements Player {

    Agent me;
    int day;
    boolean canTalk;
    boolean canWhisper;
    GameInfo currentGameInfo;
    List<Agent> aliveOthers;
    List<Agent> executedAgents = new ArrayList<>();
    List<Agent> killedAgents = new ArrayList<>();
    List<Judge> divinationList = new ArrayList<>();
    List<Judge> identList = new ArrayList<>();
    Deque<Content> talkQueue = new LinkedList<>();
    Deque<Content> whisperQueue = new LinkedList<>();
    Agent voteCandidate;
    Agent declaredVoteCandidate;
    Agent attackVoteCandidate;
    Agent declaredAttackVoteCandidate;
    Map<Agent, Role> comingoutMap = new HashMap<>();
    int talkListHead;
    List<Agent> humans = new ArrayList<>();
    List<Agent> werewolves = new ArrayList<>();
    Map<Role, Integer> roleMap;
    int updateTimes = -1;
    int gameTimes = 0;
    Map<Agent, Integer> winTimes = new HashMap<>();
    Agent firstCandidate = null;
    TalkFrequencyModel tf = TalkFrequencyModel.getInstance();
    List<Role> allRole = new ArrayList<>();
    GameSetting setting;
    List<Agent> black = new ArrayList<>();
    List<Agent> white = new ArrayList<>();
    boolean voted = false;
    boolean reVoted = false;
    int turn = -1;
    int voteDay = -2;
    Map<Agent, Integer> EstimateTime = new HashMap<>();
    HashMap<Agent, Integer> relianceList = new HashMap<>();
    Map<Agent, Boolean> coDeclaredMap = new HashMap<>();
    List<Map<Agent, Agent>> votingMapHistory = new ArrayList<>();
    //襲撃予定map
    Map<Agent, Agent> attackVotingMap;
    FakeResult fakeResult = FakeResult.getInstance();
    List<List<Talk>> allTalkList = new ArrayList<>();

    @Override
    public String getName() {
        return "DemoRoleAssignPlayer";
    }

    protected boolean isAlive(Agent agent) {
        return currentGameInfo.getStatusMap().get(agent) == Status.ALIVE;
    }

    protected boolean isKilled(Agent agent) {
        return killedAgents.contains(agent);
    }

    protected boolean isCo(Agent agent) {
        return comingoutMap.containsKey(agent);
    }

    protected boolean isCo(Role role) {
        return comingoutMap.containsKey(role);
    }

    protected boolean isHuman(Agent agent) {
        return humans.contains(agent);
    }

    protected boolean isWerewolf(Agent agent) {
        return werewolves.contains(agent);
    }

    protected <T> T randomSelect(List<T> list) {
        if (list.isEmpty()) {
            return null;
        } else {
            return list.get((int) (Math.random() * list.size()));
        }
    }

    @Override
    public void initialize(GameInfo gameInfo, GameSetting gameSetting) {
        latestGameInfo = gameInfo;
        if (gameTimes == 0) {
            for (Agent agent : gameInfo.getAgentList()) {
                winTimes.put(agent, 0);
            }
        }
        // エージェントごとの情報データ
        infoMap = new HashMap<>();
        // エージェントごとに-1（人狼）判定された回数を保存
        svmMap = new HashMap<>();
        // カミングアウトしているプレイヤーのリスト
        SeerCOAgent = new HashSet<>();
        MediumCOAgent = new HashSet<>();
        // 占い師COエージェントごとの人狼・人間判定
        wolfAgentMap = new HashMap<>();
        humanAgentMap = new HashMap<>();
        wolfAgentMapByMedium = new HashMap<>();
        humanAgentMapByMedium = new HashMap<>();
        // 占い師、霊媒師COした順番
        orderSeerCO = new HashMap<>();
        orderMediumCO = new HashMap<>();
        relianceList = new HashMap<>();
        votingMapHistory = new ArrayList<>();
        attackVotingMap = new HashMap<>();
        setting = gameSetting;
        day = -1;
        voteDay = -2;
        me = gameInfo.getAgent();
        aliveOthers = new ArrayList<>(gameInfo.getAliveAgentList());
        allTalkList = new ArrayList<>();
        aliveOthers.remove(me);
        executedAgents.clear();
        killedAgents.clear();
        divinationList.clear();
        identList.clear();
        comingoutMap.clear();
        humans.clear();
        werewolves.clear();
        roleMap = gameSetting.getRoleNumMap();
        tf.startGame(gameInfo, gameSetting);
        voted = false;
        reVoted = false;
        try {
            //学習したモデルの読み込み
            setModel();
        } catch (IOException e) {
            e.printStackTrace();
        }

        coDeclaredMap = new HashMap<>();
        for (Agent agent : gameInfo.getAgentList()) {
            coDeclaredMap.put(agent, false);
        }

        // 今日のログを取得
        List<Talk> talkList = latestGameInfo.getTalkList();
        setCO(talkList);
        // 各CO占い師,霊媒師による人狼、人間mapの更新
        setWolfHumanMap(talkList);
        // readTalkNumの更新
        readTalkNum = talkList.size();
        initDeadOrAlive();

        for (Agent agent : gameInfo.getAgentList()) {
            relianceList.put(agent, 0);
        }

    }

    public void update(GameInfo gameInfo) {
        updateTimes++;
        currentGameInfo = gameInfo;
        latestGameInfo = gameInfo;
        if (currentGameInfo.getDay() == day + 1) {
            day = currentGameInfo.getDay();
            return;
        }
        addExecutedAgent(currentGameInfo.getLatestExecutedAgent());
        for (int i = talkListHead; i < currentGameInfo.getTalkList().size(); i++) {
            Talk talk = currentGameInfo.getTalkList().get(i);
            if(allTalkList.size()!=day){
                List<Talk> list =new ArrayList<>();
                allTalkList.add(list);
            }
            allTalkList.get(day-1).add(talk);
            Agent talker = talk.getAgent();
            if (talker == me) {
                continue;
            }
            Content content = new Content(talk.getText());
            Agent subject = talk.getAgent();
            Agent target = content.getTarget();//修正
            Species species = content.getResult();//修正
            switch (content.getTopic()) {
                case COMINGOUT:
                    comingoutMap.put(talker, content.getRole());
                    TalkFrequencyModel.TopicDetailEvent tdEvent1 = new TalkFrequencyModel.TopicDetailEvent(talk.getDay(), talk.getTurn(), TalkFrequencyModel.ActionType.COMINGOUT,
                            subject);
                    tf.updateRoleProbability(tdEvent1, currentGameInfo, setting);
                    if (content.getRole() == currentGameInfo.getRole() && (content.getRole() == Role.SEER ||
                            content.getRole() == Role.MEDIUM)) {
                        for (Role role : Role.values()) {
                            if (role == Role.FOX || role == Role.FREEMASON) {
                                continue;
                            }
                            if (role != Role.POSSESSED && role != Role.WEREWOLF) {
                                tf.updateRoleProbability(subject, role, false, currentGameInfo);
                            }
                        }
                    }
                    break;
                case DIVINED:
                    divinationList.add(new Judge(day, talker, content.getTarget(), content.getResult()));
                    TalkFrequencyModel.TopicDetailEvent tdEvent = new TalkFrequencyModel.TopicDetailEvent(talk.getDay(), talk.getTurn(), TalkFrequencyModel.ActionType.DIVINED,
                            subject);
                    tf.updateRoleProbability(tdEvent, currentGameInfo, setting);
                    if (target == me) {
                        if ((currentGameInfo.getRole() != Role.WEREWOLF && species == Species.WEREWOLF) ||
                                currentGameInfo.getRole() == Role.WEREWOLF && species == Species.HUMAN) {
                            for (Role role : Role.values()) {
                                if (role == Role.FOX || role == Role.FREEMASON) {
                                    continue;
                                }
                                if (role != Role.POSSESSED && role != Role.WEREWOLF) {
                                    tf.updateRoleProbability(subject, role, false, currentGameInfo);
                                }
                            }
                        }
                    }
                    break;
                case IDENTIFIED:
                    identList.add(new Judge(day, talker, content.getTarget(), content.getResult()));
                    TalkFrequencyModel.TopicDetailEvent tdEvent2 = new TalkFrequencyModel.TopicDetailEvent(talk.getDay(), talk.getTurn(), TalkFrequencyModel.ActionType.IDENTIFIED,
                            subject);
                    tf.updateRoleProbability(tdEvent2, currentGameInfo, setting);
                    break;
                case ESTIMATE:
                    if (content.getRole() == Role.WEREWOLF || species == Species.WEREWOLF) {
                        if (votingMap.get(subject) == null && isAlive(subject)) {
                            votingMap.put(subject, target);
                        }
                        TalkFrequencyModel.TopicDetailEvent tdEvent3 = new TalkFrequencyModel.TopicDetailEvent(talk.getDay(), talk.getTurn(),
                                TalkFrequencyModel.ActionType.ESTIMATE_WOLF, subject);
                        tf.updateRoleProbability(tdEvent3, currentGameInfo, setting);
                    } else {
                        TalkFrequencyModel.TopicDetailEvent tdEvent3 = new TalkFrequencyModel.TopicDetailEvent(talk.getDay(), talk.getTurn(),
                                TalkFrequencyModel.ActionType.ESTIMATE_HUM, subject);
                        tf.updateRoleProbability(tdEvent3, currentGameInfo, setting);
                    }
                    break;
                case VOTE:
                    if (isCo(target) && comingoutMap.get(target) == Role.SEER) {
                        TalkFrequencyModel.TopicDetailEvent tdEvent3 = new TalkFrequencyModel.TopicDetailEvent(talk.getDay(), talk.getTurn(),
                                TalkFrequencyModel.ActionType.DEC_VOTE2SEER, subject);
                        tf.updateRoleProbability(tdEvent3, currentGameInfo, setting);
                    } else if (isCo(target) && comingoutMap.get(target) == Role.MEDIUM) {
                        TalkFrequencyModel.TopicDetailEvent tdEvent3 = new TalkFrequencyModel.TopicDetailEvent(talk.getDay(), talk.getTurn(),
                                TalkFrequencyModel.ActionType.DEC_VOTE2MED, subject);
                        tf.updateRoleProbability(tdEvent3, currentGameInfo, setting);
                    } else if (black.contains(target)) {
                        TalkFrequencyModel.TopicDetailEvent tdEvent3 = new TalkFrequencyModel.TopicDetailEvent(talk.getDay(), talk.getTurn(),
                                TalkFrequencyModel.ActionType.DEC_VOTE2BLACK, subject);
                        tf.updateRoleProbability(tdEvent3, currentGameInfo, setting);
                    } else if (white.contains(target)) {
                        TalkFrequencyModel.TopicDetailEvent tdEvent3 = new TalkFrequencyModel.TopicDetailEvent(talk.getDay(), talk.getTurn(),
                                TalkFrequencyModel.ActionType.DEC_VOTE2WHITE, subject);
                        tf.updateRoleProbability(tdEvent3, currentGameInfo, setting);
                    } else if (getEvil(subject)) {
                        TalkFrequencyModel.TopicDetailEvent tdEvent3 = new TalkFrequencyModel.TopicDetailEvent(talk.getDay(), talk.getTurn(),
                                TalkFrequencyModel.ActionType.DEC_VOTE2EVIL, subject);
                        tf.updateRoleProbability(tdEvent3, currentGameInfo, setting);
                    } else {
                        TalkFrequencyModel.TopicDetailEvent tdEvent3 = new TalkFrequencyModel.TopicDetailEvent(talk.getDay(), talk.getTurn(),
                                TalkFrequencyModel.ActionType.DEC_VOTE2GOOD, subject);
                        tf.updateRoleProbability(tdEvent3, currentGameInfo, setting);
                    }
                case OPERATOR:
                    Operator operator=content.getOperator();
                    if(operator != Operator.REQUEST){
                        break;
                    }
//                    if (votingMap.get(subject) == null && isAlive(subject)) {
//                        votingMap.put(subject, target);
//                    }
                default:
                    break;
            }
        }

        if (voted && day != voteDay || reVoted) {
            int count = 0;
            if (reVoted) {
                count += 1;
            }
            for (Vote vote : currentGameInfo.getLatestVoteList()) {
                if (isCo(vote.getTarget()) && comingoutMap.get(vote.getTarget()) == Role.SEER) {
                    TalkFrequencyModel.TopicDetailEvent tdEvent = new TalkFrequencyModel.TopicDetailEvent(vote.getDay(), count,
                            TalkFrequencyModel.ActionType.VOTE2SEER, vote.getAgent());
                    tf.updateRoleProbability(tdEvent, currentGameInfo, setting);
                } else if (isCo(vote.getTarget()) && comingoutMap.get(vote.getTarget()) == Role.MEDIUM) {
                    TalkFrequencyModel.TopicDetailEvent tdEvent = new TalkFrequencyModel.TopicDetailEvent(vote.getDay(), count,
                            TalkFrequencyModel.ActionType.VOTE2MED, vote.getAgent());
                    tf.updateRoleProbability(tdEvent, currentGameInfo, setting);
                } else if (isCo(vote.getTarget()) && comingoutMap.get(vote.getTarget()) == Role.MEDIUM) {
                    TalkFrequencyModel.TopicDetailEvent tdEvent = new TalkFrequencyModel.TopicDetailEvent(vote.getDay(), count,
                            TalkFrequencyModel.ActionType.VOTE2MED, vote.getAgent());
                    tf.updateRoleProbability(tdEvent, currentGameInfo, setting);
                } else if (black.contains(vote.getTarget())) {
                    TalkFrequencyModel.TopicDetailEvent tdEvent = new TalkFrequencyModel.TopicDetailEvent(vote.getDay(), count,
                            TalkFrequencyModel.ActionType.VOTE2BLACK, vote.getAgent());
                    tf.updateRoleProbability(tdEvent, currentGameInfo, setting);
                } else if (white.contains(vote.getTarget())) {
                    TalkFrequencyModel.TopicDetailEvent tdEvent = new TalkFrequencyModel.TopicDetailEvent(vote.getDay(), count,
                            TalkFrequencyModel.ActionType.VOTE2WHITE, vote.getAgent());
                    tf.updateRoleProbability(tdEvent, currentGameInfo, setting);
                } else if (getEvil(vote.getTarget())) {
                    TalkFrequencyModel.TopicDetailEvent tdEvent = new TalkFrequencyModel.TopicDetailEvent(vote.getDay(), count,
                            TalkFrequencyModel.ActionType.VOTE2EVIL, vote.getAgent());
                    tf.updateRoleProbability(tdEvent, currentGameInfo, setting);
                } else {
                    TalkFrequencyModel.TopicDetailEvent tdEvent = new TalkFrequencyModel.TopicDetailEvent(vote.getDay(), count,
                            TalkFrequencyModel.ActionType.VOTE2GOOD, vote.getAgent());
                    tf.updateRoleProbability(tdEvent, currentGameInfo, setting);
                }
            }
            voteDay = day;
            reVoted = false;
        }

        setReliance(currentGameInfo.getTalkList());

        talkListHead = currentGameInfo.getTalkList().size();


    }

    //votingMap(投票予定先)の更新
    private void setReliance(List<Talk> talkList) {
        if (latestGameInfo.getAliveAgentList().contains(
                latestGameInfo.getAgent())) {
            for (int i = readTalkNum; i < talkList.size(); i++) {
                Talk talk = talkList.get(i);
                for (int j = 0; j < i; j++) {
                    Talk oldTalk = talkList.get(j);
                    if (talk.getAgent() == oldTalk.getAgent() && talk.getText() == oldTalk.getText()) {
                        return;
                    }
                }
                // 発話をパース
                Content content = new Content(talk.getText());//修正

                Agent subject = talk.getAgent();
                Agent target = content.getTarget();//修正

                if(subject==null||target==null||!relianceList.containsKey(subject)||!relianceList.containsKey(target)){
                    continue;
                }

                //double reliance = ((relianceList.get(subject) / 500.0) - 1.0) * 0.1 + 1.0;
                switch (content.getTopic()) {
                    //何回も投票先を変更する相手のために変更の制限回数を設ける
                    case VOTE:
                        if (latestGameInfo.getAliveAgentList().contains(target)) {// vote対象が生きていないなら意味ない
                            if (votingMap.get(subject) != target) {
                                votingMap.put(subject, target);
                                int temp = relianceList.get(target);
                                relianceList.put(target, temp - 300);
                            }
                        }
                        break;
                    case COMINGOUT:
                        Role COrole = content.getRole();//修正
                        if (subject == target) {
                            int temp = 0;
                            switch (COrole) {
                                case BODYGUARD:
                                    if (!coDeclaredMap.get(subject)) {
                                        temp = relianceList.get(target);
                                        relianceList.put(target, temp + 200);
                                        coDeclaredMap.put(target, true);
                                    }
                                    break;
                                case SEER:
                                    if (!coDeclaredMap.get(subject)) {
                                        temp = relianceList.get(target);
                                        relianceList.put(target, temp + 300);
                                        coDeclaredMap.put(target, true);
                                    }

                                    break;
                                case MEDIUM:
                                    if (!coDeclaredMap.get(subject)) {
                                        temp = relianceList.get(target);
                                        relianceList.put(target, temp + 200);
                                        coDeclaredMap.put(target, true);
                                    }
                                    break;
                                case POSSESSED:
                                default:
                                    break;
                            }
                        }
                        break;
                    case ESTIMATE:
                        int temp;
                        if (EstimateTime.get(subject) < 3) {
                            switch (content.getRole()) {
                                case VILLAGER:
                                    temp = relianceList.get(target);
                                    relianceList.put(target, temp + 50);
                                    break;
                                case BODYGUARD:
                                    temp = relianceList.get(target);
                                    relianceList.put(target, temp + 50);
                                    break;
                                case SEER:
                                    temp = relianceList.get(target);
                                    relianceList.put(target, temp + 100);
                                    break;
                                case MEDIUM:
                                    temp = relianceList.get(target);
                                    relianceList.put(target, temp + 50);
                                    break;
                                case POSSESSED:
                                    temp = relianceList.get(target);
                                    relianceList.put(target, temp - 50);
                                    break;
                                case WEREWOLF:
                                    temp = relianceList.get(target);
                                    relianceList.put(target, temp - 100);
                                    break;
                                default:
                                    break;
                            }
                            EstimateTime.put(subject, EstimateTime.get(subject) + 1);
                        }
                        break;
                    case DIVINED:
                        temp = relianceList.get(target);
                        if (content.getResult() == Species.HUMAN) {
                            relianceList.put(target, temp + 300);
                        } else {
                            relianceList.put(target, temp - 300);
                        }
                        break;
//                    case AGREE:
//                        int talkId = content.getTalkID();
//                        int talkDay= content.getTalkDay();
//                        Talk targetTalk = allTalkList.get(talkDay-1).get(talkId);
//                        Agent talkedAgent = targetTalk.getAgent();
//                        temp = relianceList.get(talkedAgent);
//                        relianceList.put(talkedAgent, temp + 100);
//                        break;
//                    case DISAGREE:
//                        talkId = content.getTalkID();
//                        talkDay = content.getTalkDay();
//                        targetTalk = allTalkList.get(talkDay-1).get(talkId);
//                        talkedAgent = targetTalk.getAgent();
//                        temp = relianceList.get(talkedAgent);
//                        relianceList.put(talkedAgent, temp - 100);
//                        break;
                    default:
                        break;
                }

            }
        }
    }


    private boolean getEvil(Agent subject) {
        // TODO 自動生成されたメソッド・スタブ
        Map<Agent, Double> evilList = new HashMap<Agent, Double>();

        for (Agent agent : currentGameInfo.getAliveAgentList()) {
            evilList.put(agent, tf.getRoleProbability(agent, Role.WEREWOLF));
        }

        evilList.put(subject, tf.getRoleProbability(subject, Role.WEREWOLF));

        for (Agent seerAgent : SeerCOAgent) {
            for (Agent wolf : wolfAgentMap.get(seerAgent)) {
                double tmp = evilList.get(wolf) + (0.3 + 3 * tf.getRoleProbability(seerAgent, Role.SEER));
                evilList.put(wolf, tmp);
            }
            for (Agent human : humanAgentMap.get(seerAgent)) {
                double tmp = evilList.get(human) - (0.3 + 3 * tf.getRoleProbability(seerAgent, Role.SEER));
                evilList.put(human, tmp);
            }

        }

        for (Agent agent : evilList.keySet()) {
            if (isCo(agent)) {
                if (comingoutMap.get(agent) == Role.SEER && day <= 4) {
                    evilList.put(agent, 0.0);
                    continue;
                } else if (comingoutMap.get(agent) == Role.MEDIUM && day <= 3) {
                    evilList.put(agent, 0.0);
                    continue;
                }
            }
            if (agent == me & currentGameInfo.getRole() != Role.WEREWOLF) {
                evilList.put(agent, 0.0);
                continue;
            }

            if (evilList.get(agent) < 0) {
                evilList.put(agent, 0.0);
            }
        }

        double mean = 0.0;

        for (Agent agent : evilList.keySet()) {
            mean += evilList.get(agent);
        }

        mean /= evilList.size();
        if (mean <= evilList.get(subject)) {
            return true;
        } else {
            return false;
        }
    }

    @Override
    public void dayStart() {
        updateTimes = -1;
        turn = -1;
        canTalk = true;
        canWhisper = false;
        if (currentGameInfo.getRole() == Role.WEREWOLF) {
            canWhisper = true;
        }

        talkQueue.clear();
        whisperQueue.clear();
        declaredVoteCandidate = null;
        voteCandidate = null;
        declaredAttackVoteCandidate = null;
        attackVoteCandidate = null;
        talkListHead = 0;
        voted = false;
        reVoted = false;
        //特徴量の取得
        getInfo();
        if (latestGameInfo.getDay() > 0) {
            try {
                setSVM();
            } catch (Exception e) {
                e.printStackTrace();
            }
        }
        // 各エージェントの生死の更新
        setDeadOrAlive();
        addExecutedAgent(currentGameInfo.getExecutedAgent());
        if (!currentGameInfo.getLastDeadAgentList().isEmpty()) {
            addKilledAgent(currentGameInfo.getLastDeadAgentList().get(0));
        }

        for (Agent agent : currentGameInfo.getLastDeadAgentList()) {
            tf.updateRoleProbability(agent, Role.WEREWOLF, false, currentGameInfo);
            for (Judge judge : divinationList) {
                if (judge.getResult() == Species.WEREWOLF) {
                    for (Role role : Role.values()) {
                        if (role == Role.FOX || role == Role.FREEMASON) {
                            continue;
                        }
                        if (role != Role.POSSESSED && role != Role.WEREWOLF) {
                            tf.updateRoleProbability(judge.getAgent(), role, false, currentGameInfo);
                        }
                    }
                }
            }
        }

        for (Agent agent : latestGameInfo.getAgentList()) {
            relianceList.put(agent, 0);
        }

        if (day > 0 && latestGameInfo.getRole()!=Role.WEREWOLF) {
            Map<Agent, Agent> temp = new HashMap<>();
            for (Vote vote : latestGameInfo.getVoteList()) {
                if(!isAlive(vote.getTarget())&&vote.getTarget()!=latestGameInfo.getLatestExecutedAgent()){
                    continue;
                }
                temp.put(vote.getAgent(), vote.getTarget());
            }
            votingMapHistory.add(temp);

        }
        votingMap = new HashMap<>();

        if (day > 1) {
            for (Vote vote : latestGameInfo.getVoteList()) {
                if (vote.getTarget() != null && isAlive(vote.getAgent()) && isAlive(vote.getTarget())) {
                    int reliance = relianceList.get(vote.getTarget());
                    relianceList.put(vote.getTarget(), reliance - 300);
                }
            }
        }

        for (Agent agt : latestGameInfo.getAliveAgentList()) {
            votingMap.put(agt, null);
        }
        attackVoteCandidate = null;
        attackVotingMap = new HashMap<>();
        for (Agent agt : latestGameInfo.getAliveAgentList()) {
            if (werewolves.contains(agt)) {
                votingMap.put(agt, null);
            }
        }
        for (Agent agent : latestGameInfo.getAgentList()) {
            EstimateTime.put(agent, 0);
        }
    }

    public void addKilledAgent(Agent killedAgent) {
        if (killedAgent != null) {
            aliveOthers.remove(killedAgent);
            if (!killedAgents.contains(killedAgent)) {
                killedAgents.add(killedAgent);
            }
        }
    }

    public void addExecutedAgent(Agent executedAgent) {
        if (executedAgent != null) {
            aliveOthers.remove(executedAgent);
            if (!executedAgents.contains(executedAgent)) {
                executedAgents.add(executedAgent);
            }
        }
    }

    public void chooseVoteCandidate() {

    }

    @Override
    public String talk() {
        turn += 1;

        if (voteCandidate != null && voteCandidate != declaredVoteCandidate) {
            talkQueue.offer(new Content(new VoteContentBuilder(voteCandidate)));
            declaredVoteCandidate = voteCandidate;
        }

        return talkQueue.isEmpty() ? Talk.SKIP : talkQueue.poll().getText();
    }

    public void chooseAttackVoteCandidate() {

    }

    public String whisper() {
        chooseAttackVoteCandidate();
        if (attackVoteCandidate != null && attackVoteCandidate != declaredAttackVoteCandidate
                && day != 0) {
            whisperQueue.offer(new Content(new AttackContentBuilder(attackVoteCandidate)));
            declaredAttackVoteCandidate = attackVoteCandidate;
        }
        if (declaredAttackVoteCandidate != null) {
            whisperQueue.offer(new Content(new RequestContentBuilder(null,
                    new Content(new AttackContentBuilder(declaredAttackVoteCandidate)))));
        }
        return whisperQueue.isEmpty() ? Talk.SKIP : whisperQueue.poll().getText();
    }

    public Agent vote() {
        voted = true;
        if (voted) {
            reVoted = true;
        }
        return voteCandidate;
    }

    public Agent attack() {
        canWhisper = false;
        chooseAttackVoteCandidate();
        canWhisper = true;
        return attackVoteCandidate;
    }

    public Agent divine() {
        return randomSelect(aliveOthers);
    }

    public Agent guard() {
        return null;
    }

    public void finish() {
        AnalysisOfGuess(currentGameInfo);
    }

    public void AnalysisOfGuess(GameInfo gameInfo) {
        gameTimes++;

        tf.endGame(gameInfo);
        int result = 0;
        for (Agent agent : currentGameInfo.getAliveAgentList()) {
            if (currentGameInfo.getRoleMap().get(agent) == Role.WEREWOLF) {
                result = 1;
                break;
            }
        }
        for (Agent agent : currentGameInfo.getAgentList()) {
            int win = winTimes.get(agent);
            if (currentGameInfo.getRoleMap().get(agent) == Role.WEREWOLF) {
                winTimes.put(agent, win + result);
            } else {
                winTimes.put(agent, win + 1 - result);
            }
        }

//		for (Agent agent : currentGameInfo.getAgentList()) {
//			if (agent == me) {
//				continue;
//			}
//			if (currentGameInfo.getRole() != Role.WEREWOLF) {
//				System.out.print(agent.getAgentIdx() + " 実際の役職:" + currentGameInfo.getRoleMap().get(agent) + " 人狼の確率:"
//						+ tf.getRoleProbability(agent, Role.WEREWOLF) + "\n");
//			} else {
//				System.out.print(agent.getAgentIdx() + " 実際の役職:" + currentGameInfo.getRoleMap().get(agent) + " 狩人の確率:"
//						+ tf.getRoleProbability(agent, Role.BODYGUARD) + "\n");
//			}
//		}

        int win = -1;
        firstCandidate = null;
        for (Agent agent : currentGameInfo.getAgentList()) {
            if (agent == me) {
                continue;
            }
            if (win < winTimes.get(agent)) {
                win = winTimes.get(agent);
                firstCandidate = agent;
            }
        }
        fakeResult.Guess(gameInfo, gameTimes, Role.VILLAGER, false);
    }

    //ゲーム情報
    public GameInfo latestGameInfo;
    // エージェントごとの情報データ
    Map<Agent, Integer[]> infoMap = new HashMap<>();
    // エージェントごとに-1（人狼）判定された回数を保存
    Map<Agent, Integer> svmMap = new HashMap<>();
    // 投票予定map
    Map<Agent, Agent> votingMap;

    private void getInfo() {
        if (latestGameInfo.getDay() > 0) {

            for (Agent agent : latestGameInfo.getAgentList()) {

                Integer[] info = new Integer[11];
                //日にち　すべてのiで共通 status
                info[10] = latestGameInfo.getDay() - 1;
                //生きてるなら０、死んでるなら１ status
                if (AliveOrDeadMap.get(agent) == Status.ALIVE) {
                    info[0] = 0;
                } else {
                    info[0] = 1;
                }
                //CO占い師数 すべてのiで共通 talk
                info[1] = SeerCOAgent.size();
                //CO霊媒師数　すべてのiで共通 talk
                info[2] = MediumCOAgent.size();
                //受けた占い人間判定数 talk
                info[3] = 0;
                //受けた占い人狼判定数 talk
                info[4] = 0;
                for (Agent seer : SeerCOAgent) {
                    if (humanAgentMap.get(seer).contains(agent)) {
                        info[3]++;
                    }
                    if (wolfAgentMap.get(seer).contains(agent)) {
                        info[4]++;
                    }
                }
                //VOTE発話から投票変更 talk & vote
                if (!latestGameInfo.getVoteList().isEmpty()) {
                    if (AliveOrDeadMap.get(agent) == Status.ALIVE) {
                        for (int i = 0; i < latestGameInfo.getVoteList().size(); i++) {
                            if (latestGameInfo.getVoteList().get(i).getAgent() == agent) {
                                //投票変更したなら
                                if (latestGameInfo.getVoteList().get(i)
                                        .getTarget() != votingMap.get(agent)) {
                                    info[9] = infoMap.get(agent)[9] + 1;
                                } else {
                                    info[9] = infoMap.get(agent)[9];
                                }
                            }
                        }
                    } else {
                        info[9] = infoMap.get(agent)[9];
                    }

                } else {
                    info[9] = 0;
                }

                //何番目に占い師CO talk
                info[5] = 0;
                //何番目に霊媒師CO talk
                info[6] = 0;
                //占い師で人間判定を出した数 talk
                info[7] = 0;
                //占い師で人狼判定を出した数 talk
                info[8] = 0;
                if (SeerCOAgent.contains(agent)) {
                    info[5] = orderSeerCO.get(agent);
                    info[7] = humanAgentMap.get(agent).size();
                    info[8] = wolfAgentMap.get(agent).size();
                }

                if (MediumCOAgent.contains(agent)) {
                    info[6] = orderMediumCO.get(agent);
                }
                //infoMapを更新
                infoMap.put(agent, info);
            }

        }
    }

    // モデルのロード
    svm_model model;

    private void setModel() throws IOException {
//        model = svm.svm_load_model("log.model");
        model=svm.svm_load_model(new InputStreamReader(DemoBasePlayer.class.getClassLoader().getResourceAsStream("submit/aiwolf/wordwolf/log.model")));
    }

    //対戦相手の判別
    private void setSVM() throws Exception {

        for (Agent agt : latestGameInfo.getAgentList()) {

            int node = 11;
            black.clear();
            white.clear();

            // 判別対象にしたいデータ
            svm_node[] input = new svm_node[node];
            for (int i = 0; i < node; i++) {
                input[i] = new svm_node();
            }
            // ラベルをセット
            for (int i = 0; i < node; i++) {
                input[i].index = i + 1;
            }

            // 値をセット
            for (int j = 0; j < node; j++) {
                if (Objects.toString(infoMap.get(agt)[j]) != "null") {
                    input[j].value = infoMap.get(agt)[j];
                    //System.out.println("成功" + input[j].value);
                } else {

                }
            }

            // 判別の実行
            double v = 0;
            try {
                v = svm.svm_predict(model, input);
            } catch (Exception e) {
                //System.out.println("判別エラー");
            }

            if (v == -1) {
                //System.out.println("人狼判定");
                black.add(agt);
                Integer in = svmMap.get(agt);
                int i;
                if (in == null) {
                    i = 0;
                } else {
                    i = in.intValue();
                }
                i++;
                in = Integer.valueOf(i);
                svmMap.put(agt, in);
            }
            if (v == 1.0) {
                white.add(agt);
            }
        }
    }

    // 各エージェントの生死
    Map<Agent, Status> AliveOrDeadMap = new HashMap<Agent, Status>();

    private void initDeadOrAlive() {
        for (Agent agt : latestGameInfo.getAgentList()) {
            AliveOrDeadMap.put(agt, Status.ALIVE);
        }
    }

    private void setDeadOrAlive() {
        Agent excutedAgent = latestGameInfo.getExecutedAgent();
        if (excutedAgent != null) {
            AliveOrDeadMap.put(excutedAgent, Status.DEAD);
        }
        List<Agent> deadAgentList = latestGameInfo.getLastDeadAgentList();//修正
        for (Agent agt : deadAgentList) {
            AliveOrDeadMap.put(agt, Status.DEAD);
        }
    }

    // カミングアウトしているプレイヤーのリスト
    Set<Agent> SeerCOAgent = new HashSet<Agent>();
    Set<Agent> MediumCOAgent = new HashSet<Agent>();
    // 占い師COエージェントごとの人狼・人間判定
    Map<Agent, Set<Agent>> wolfAgentMap = new HashMap<Agent, Set<Agent>>();
    Map<Agent, Set<Agent>> humanAgentMap = new HashMap<Agent, Set<Agent>>();
    Map<Agent, Set<Agent>> wolfAgentMapByMedium = new HashMap<Agent, Set<Agent>>();
    Map<Agent, Set<Agent>> humanAgentMapByMedium = new HashMap<Agent, Set<Agent>>();
    // 占い師、霊媒師COした順番
    Map<Agent, Integer> orderSeerCO = new HashMap<>();
    Map<Agent, Integer> orderMediumCO = new HashMap<>();
    // その日のログの何番目まで読み込んだか
    int readTalkNum;

    private void setCO(List<Talk> talkList) {
        //自分が生きていれば
        if (latestGameInfo.getAliveAgentList().contains(
                latestGameInfo.getAgent())) {
            for (int i = readTalkNum; i < talkList.size(); i++) {
                Talk talk = talkList.get(i);
                // 発話をパース
                Content content = new Content(talk.getText());//修正

                Agent subject = talk.getAgent();
                Agent target = content.getTarget();//修正
                Role COrole = content.getRole();//修正
                Set<Agent> wolf = new HashSet<Agent>();
                Set<Agent> human = new HashSet<Agent>();

                switch (content.getTopic()) {//修正
                    case COMINGOUT:
                        // targetとsubjectが異なっていたら、そのCOは意味がない
                        if (target == subject) {
                            switch (COrole) {
                                case SEER:
                                    SeerCOAgent.add(target);
                                    wolfAgentMap.put(subject, wolf);
                                    humanAgentMap.put(subject, human);
                                    orderSeerCO.put(subject, SeerCOAgent.size());
                                    break;
                                case MEDIUM:
                                    MediumCOAgent.add(target);
                                    wolfAgentMapByMedium.put(subject, wolf);
                                    humanAgentMapByMedium.put(subject, human);
                                    orderMediumCO.put(subject, MediumCOAgent.size());
                                    break;
                                default:
                                    break;
                            }
                        }
                        break;
                    // COしていなくても、発言からCOと見なせるときはCOとみなす
                    case DIVINED:
                        SeerCOAgent.add(subject);
                        humanAgentMap.put(subject, human);
                        wolfAgentMap.put(subject, wolf);
                        break;
                    case IDENTIFIED://修正
                        MediumCOAgent.add(subject);
                        wolfAgentMapByMedium.put(subject, wolf);
                        humanAgentMapByMedium.put(subject, human);
                        break;
                    default:
                        break;
                }
            }
        }
    }

    //判定結果の更新
    private void setWolfHumanMap(List<Talk> talkList) {

        if (latestGameInfo.getDay() > 0
                && latestGameInfo.getAliveAgentList().contains(
                latestGameInfo.getAgent())) {
            for (int i = readTalkNum; i < talkList.size(); i++) {
                Talk talk = talkList.get(i);
                // 発話をパース
                Content content = new Content(talk.getText());//修正

                Agent subject = talk.getAgent();
                Agent target = content.getTarget();//修正
                Species species = content.getResult();//修正

                if (content.getTopic() == Topic.DIVINED) {//修正

                    if (species == Species.WEREWOLF) {
                        Set<Agent> wolfAgent = wolfAgentMap.get(subject);
                        wolfAgent.add(target);
                        wolfAgentMap.put(subject, wolfAgent);
                    } else if (species == Species.HUMAN) {
                        Set<Agent> humanAgent = humanAgentMap.get(subject);
                        humanAgent.add(target);
                        humanAgentMap.put(subject, humanAgent);
                    }

                } else if (content.getTopic() == Topic.IDENTIFIED) {//修正

                    if (species == Species.WEREWOLF) {
                        Set<Agent> wolfAgent = wolfAgentMapByMedium
                                .get(subject);
                        wolfAgent.add(target);
                        wolfAgentMapByMedium.put(subject, wolfAgent);
                    } else if (species == Species.HUMAN) {
                        Set<Agent> humanAgent = humanAgentMapByMedium
                                .get(subject);
                        humanAgent.add(target);
                        humanAgentMapByMedium.put(subject, humanAgent);
                    }
                }
            }
        }
    }
}
