package submit.aiwolf.wordwolf;

import java.io.BufferedReader;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.io.InputStreamReader;

import org.aiwolf.common.data.Agent;
import org.aiwolf.common.data.Judge;
import org.aiwolf.common.data.Role;
import org.aiwolf.common.data.Species;
import org.aiwolf.common.data.Talk;
import org.aiwolf.common.data.Vote;
import org.aiwolf.common.net.GameInfo;
import org.aiwolf.common.net.GameSetting;


public class TalkFrequencyModel{

	private static TalkFrequencyModel instance= new TalkFrequencyModel();

	  private TalkFrequencyModel() {
	    // 最初はインスタンスを生成しない
		  instance=null;
	  }

	  public static TalkFrequencyModel getInstance() {
	    // メソッド呼び出しがあったときに、
	    // 初めてインスタンスを生成
	    if (instance == null) {
	    	instance = new TalkFrequencyModel();
	    }
	    return instance;
	  }

	public static class TopicDetailEvent {

        int day;
        int turn;
        ActionType type;
        int initiatorIndex;
        Agent agent;

        @Override
        public String toString() {
            return day + "\t" + turn + "\t" + type;
        }

        public TopicDetailEvent(int day, int turn, ActionType type, Agent agent) {
            this.day = Math.min(day, 7);
            this.turn = Math.min(turn, 10);
            this.type = type;
            this.initiatorIndex = agent.getAgentIdx();
            this.agent=agent;
        }

    }

    public enum ActionType {

        OVER,
        DEC_VOTE2SEER,
        DEC_VOTE2MED,
        DEC_VOTE2BLACK,
        DEC_VOTE2WHITE,
        DEC_VOTE2EVIL,
        DEC_VOTE2GOOD,
        ESTIMATE_WOLF,
        ESTIMATE_HUM,
        VOTE2SEER,
        VOTE2MED,
        VOTE2BLACK,
        VOTE2WHITE,
        VOTE2EVIL,
        VOTE2GOOD,
        DIVINED,
        IDENTIFIED,
        COMINGOUT,
        OPERATOR
    }

    /* agent の role が r である 確率 */
    private Map<Agent, Map<Role, Double>> roleProbability;
    private Map<Agent,BaiseAgent> baiseAgentMap=new HashMap<>();
    private Map<Agent, Map<Integer, Map<Integer, Map<Role, Map<ActionType, Double>>>>> actConditionalProbability = new HashMap<>();


    /* 各役職の人数 */
    private Map<Role, Double> roleNum;

    public class BaiseAgent{
    	Role role;
    	Agent agent;
    }

    public void startGame(GameInfo gameInfo,GameSetting gameSetting) {
        /* 役職人数を初期化 */
        roleNum = new HashMap<>();
        baiseAgentMap.clear();
        if (gameSetting.getPlayerNum() == 5) {
            roleNum.put(Role.VILLAGER, 2.0);
            roleNum.put(Role.SEER, 1.0);
            roleNum.put(Role.WEREWOLF, 1.0);
            roleNum.put(Role.POSSESSED, 1.0);
        } else {
            roleNum.put(Role.VILLAGER, 8.0);
            roleNum.put(Role.SEER, 1.0);
            roleNum.put(Role.MEDIUM, 1.0);
            roleNum.put(Role.BODYGUARD, 1.0);
            roleNum.put(Role.WEREWOLF, 3.0);
            roleNum.put(Role.POSSESSED, 1.0);
        }
        /* エージェント別役職確率を初期化 */
        roleProbability = new HashMap<>();
        Role myRole = gameInfo.getRole();
        for (Agent ga : gameInfo.getAgentList()) {
        	BaiseAgent agent=new BaiseAgent();
        	agent.agent=ga;
        	if(gameInfo.getRole() == Role.WEREWOLF) {
        		for(Agent wolfAgent:gameInfo.getRoleMap().keySet()){
        			if(agent.agent==wolfAgent) {
        				agent.role=Role.WEREWOLF;
        			}
        		}
        	}
        	if(ga==gameInfo.getAgent()) {
        		agent.role=myRole;
        	}
        	baiseAgentMap.put(ga, agent);
            roleProbability.put(ga, new HashMap<>());
            Map<Role, Double> rp = roleProbability.get(ga);
            if (ga == gameInfo.getAgent() || (myRole == Role.WEREWOLF && agent.role == Role.WEREWOLF)) {
                roleNum.keySet().stream().forEach(x -> {
                    if (x == agent.role) {
                        rp.put(x, 1.0);
                    } else {
                        rp.put(x, 0.0);
                    }
                });
            } else {
                int vil = gameInfo.getAgentList().size() - (myRole == Role.WEREWOLF && gameInfo.getAgentList().size() == 15 ? 3 : 1);
                roleNum.keySet().stream().forEach(x -> {
                    double num = roleNum.get(x);
                    if (x == myRole) {
                        num -= myRole == Role.WEREWOLF && gameInfo.getAgentList().size() == 15 ? 3.0 : 1.0;
                    }
                    rp.put(x, num / vil);
                });
            }
            /* 条件付き確率の初期化 */
            actConditionalProbability.put(ga, getTopicProbabilityByIndex(ga.getAgentIdx()));
        }
    }



    public class Event{
    	EventType type;
    	/**
         * イベントの起きた日
         */
        public int day;
        /**
         * イベントの対象エージェントを返す（ATTACK/EXECUTE/DIVINE/MEDIUM/VICTIM_DECIDED/GUARD;ATTACK/EXECUTE以外は当該役職以外の人には判らない）。
         */
        public BaiseAgent target;
        /**
         * 判定結果（DIVINE/MEDIUM）。
         */
        public Species species;
        /**
         * 投票内容（VOTE/ATTACK_VOTE）
         */
        public List<Vote> votes;
        /**
         * 会話内容（TALK/WHISPER）
         */
        public List<Talk> talks;

        Event(EventType type, int day) {
            this.type = type;
            this.day = day;
        }

        Event(EventType type, int day, List<Vote> votes) {
            this.type = type;
            this.day = day;
            this.votes = votes;
        }

        Event(EventType type, List<Talk> talks) {
            this.type = type;
            this.day = talks.get(0).getDay();
            this.talks = talks;
        }

        Event(EventType type, Judge judge) {
            this.type = type;
            this.day = judge.getDay() - 1;
            this.species = judge.getResult();
            this.target =baiseAgentMap.get(judge.getTarget());
        }

        Event(EventType type, int day, BaiseAgent target) {
            this.type = type;
            this.day = day;
            this.target = target;
        }

    }

    public void handleEvent(GameInfo gameInfo, Event e) {

    }

    /**
     * target が role で ある/ない ことが確定した場合に確率を更新する。
     *
     * @param agent
     * @param role 是非の確定した役職
     * @param b ある(true)/ない(false)のどちらであるか
     * @param gameInfo
     */
    void updateRoleProbability(Agent agent, Role role, boolean b,GameInfo gameInfo) {

        BaiseAgent target=new BaiseAgent();
    	target.agent=agent;
    	target.role=role;

    	if (agent == gameInfo.getAgent()) {
            return;
        }

        if (b) {
            for (Map.Entry<Role, Double> entry : roleProbability.get(agent).entrySet()) {
                entry.setValue(0.0);
            }
            roleProbability.get(agent).replace(role, 1.0);
            baiseAgentMap.put(agent, target);
        } else {
            roleProbability.get(agent).put(role, 0.0);
            double sum = roleProbability.get(agent).values().stream().mapToDouble(d -> d).sum();
            for (Map.Entry<Role, Double> entry : roleProbability.get(agent).entrySet()) {
                entry.setValue(entry.getValue() / sum);
            }
        }
    }

    private List<TopicDetailEvent> events = new ArrayList<>();

    /**
     * agent のday日目第turnターンの発言のトピックが topic だった時 ベイズの公式に従って確率を更新する。 （事後確率を次の事前確率として使う） P_new (role) = P(role | talk) = (P(talk | role) * P(role))/Sum_for_roles(P(talk | role) * P(role))
     *
     * @param event
     * @param gameInfo
     * @param gameSetting
     */
    void updateRoleProbability(TopicDetailEvent event,GameInfo gameInfo,GameSetting gameSetting) {
        Agent agent= event.agent;
        events.add(event);
        if (agent == gameInfo.getAgent()) {
            return;
        }
        Map<Role, Double> newP = new HashMap<>();
        double z = 0.0; // 規格化定数
        for (Entry<Role, Double> entry : roleProbability.get(agent).entrySet()) {
            final Role role = entry.getKey();
            final double p = entry.getValue();
            newP.put(role, p * getTopicConditionalProbability(agent, event.day, event.turn, role, event.type,gameSetting));
            z += newP.get(role);
        }
        for (Entry<Role, Double> entry : roleProbability.get(agent).entrySet()) {
            Role role = entry.getKey();
            entry.setValue(newP.get(role) / z);
        }
    }

    private double getTopicConditionalProbability(Agent agent, int day, int turn, Role role, ActionType topic,GameSetting gameSetting) {
        Map<Integer, Map<Integer, Map<Role, Map<ActionType, Double>>>> tcp = actConditionalProbability.get(agent);
        if (tcp.get(day) == null) {
            tcp.put(day, new HashMap<>());
        }
        if (tcp.get(day).get(turn) == null) {
            tcp.get(day).put(turn, new HashMap<>());
            for (Role r : Role.values()) {
                if (role == Role.FOX || role == Role.FREEMASON) {
                    continue;
                }
                tcp.get(day).get(turn).put(r, getCommonTCP(day, turn, r,gameSetting));
            }
        }
        return actConditionalProbability.get(agent).get(day).get(turn).get(role).get(topic);
    }

    private Map<String, Double> P;

    private Map<ActionType, Double> getCommonTCP(int day, int turn, Role role,GameSetting gameSetting) {
        /* ファイルから事前分布を読む */
        if (P == null) {
            P = new HashMap<>();
            String filename = gameSetting.getPlayerNum() == 15 ? "actFrequency15.dat" : "actFrequency5.dat";
            try (BufferedReader br = new BufferedReader(new InputStreamReader(TalkFrequencyModel.class.getClassLoader().getResourceAsStream("submit/aiwolf/wordwolf/" + filename)))) {;
            //try (BufferedReader br = new BufferedReader(new FileReader(filename))) {;
                String line = null;
                while ((line = br.readLine()) != null) {
                    String[] data = line.split("\t");
                    P.put(data[0], Double.parseDouble(data[1]));
                }
            } catch (Exception e) {
                e.printStackTrace();
            }
        }
        Map<ActionType, Double> ret = new HashMap<>();
        for (ActionType t : ActionType.values()) {
            ret.put(t, P.getOrDefault(String.join(",", "" + role, "" + day, "" + turn, "" + t), 1.0));
        }
        /* 規格化定数 */
        final double z = ret.values().stream().mapToDouble(x -> x).sum();
        ret.entrySet().stream().forEach(x -> x.setValue(x.getValue() / z));
        return ret;
    }

    public double getRoleProbability(Agent agent, Role role) {
        return roleProbability.get(agent).get(role);
    }

    public Map<Role, Double> getRoleProbability(Agent agent) {
        return roleProbability.get(agent);
    }

    public Map<Agent, Map<Role, Double>> getRoleProbability() {
        return roleProbability;
    }

    //metagame要素
    public HashCounter<String> evc = new HashCounter<>();


    public Map<Integer, Map<Integer, Map<Role, Map<ActionType, Double>>>> getTopicProbabilityByIndex(int index) {
        Map<Integer, Map<Integer, Map<Integer, Map<Role, Map<ActionType, Double>>>>> tp = metaActProbability;
        if (tp.get(index) == null) {
            tp.put(index, new HashMap<>());
        }
        return tp.get(index);
    }

    /* エージェント別役職経験回数 */
    private Map<Integer, Map<Role, Integer>> roleCount = new HashMap<>();

    /**
     * 1ゲーム終了時、そのゲーム中発生した発言に対応する確率を、少し上げる
     *
     * @param g
     */
    private void updateTopicProbability(GameInfo g) {

    	for(Entry<Agent,Role>roleMap:g.getRoleMap().entrySet()) {
        	if (!roleCount.containsKey(roleMap.getKey().getAgentIdx())) {
                roleCount.put(roleMap.getKey().getAgentIdx(), new HashMap<>());
            }
            if (roleCount.get(roleMap.getKey().getAgentIdx()).containsKey(roleMap.getValue())) {
                roleCount.get(roleMap.getKey().getAgentIdx()).put(roleMap.getValue(), roleCount.get(roleMap.getKey().getAgentIdx()).get(roleMap.getValue()) + 1);
            } else {
                roleCount.get(roleMap.getKey().getAgentIdx()).put(roleMap.getValue(), 1);
            }
        }

        events.forEach(ev -> {
            Agent ga = ev.agent;
            if (g.getAgent() != ga) {
                final int day = ev.day;
                final int turn = ev.turn;
                final int agentIndex = ga.getAgentIdx();
                final Role role = g.getRoleMap().get(ga);
                final ActionType topic = ev.type;
                evc.countPlus(String.join("\t", "" + agentIndex, "" + day, "" + turn, "" + role, "" + topic));
                for (Entry<ActionType, Double> entry : metaActProbability.get(agentIndex).get(day).get(turn).get(role).entrySet()) {
//                    double freq = entry.getValue() * (23.0 + g.getMeta().numberOfGames - 1.0);
//                    double div = 23.0 + g.getMeta().numberOfGames;
                    double freq = entry.getValue() * (12.0 + roleCount.get(agentIndex).get(role));
                    double div = 13.0 + roleCount.get(agentIndex).get(role);
                    entry.setValue(entry.getKey() == topic ? (freq + 1.0) / div : freq / div);
                }
            }
        });
    }

    public void endGame(GameInfo gameInfo) {
        updateTopicProbability(gameInfo);
        events.clear();
    }


    private Map<Integer, Map<Integer, Map<Integer, Map<Role, Map<ActionType, Double>>>>> metaActProbability = new HashMap<>();
}
